import re

import docx
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_BREAK
from docx.enum.style import WD_STYLE_TYPE
from docx.shared import Pt, Cm

from md2book.util.common import get_file_local_path
from .mdhtml import purify_remove_html

ALIGMENTS = {
	'left' : WD_ALIGN_PARAGRAPH.LEFT,
	'right' : WD_ALIGN_PARAGRAPH.RIGHT,
	'center' : WD_ALIGN_PARAGRAPH.CENTER,
	'justify' : WD_ALIGN_PARAGRAPH.JUSTIFY,
}
REGEX_IMAGES = r'\[\[IMAGE-ITEM\]\]\[\[(.*?)\]\]\[\[(.*?)\]\]'
SEP_MARGIN = Pt(20)

def create_document_styles(document):
	sep_style = document.styles.add_style('SepStyle', WD_STYLE_TYPE.PARAGRAPH)
	sep_style.base_style = document.styles['Body Text']
	# sep_style.font.size = docx.shared.Pt(14)

def post_process_docx(docx_file, target):
	docx_path = docx_file.getStrPath()
	document = docx.Document(docx_path)
	create_document_styles(document)

	alignment = target['style']['align']
	sep = target['sep']
	heading_sep = 'Heading ' + str(target['chapter-level'])
	image_alignment = 'center' if target['style']['center-blocks'] else alignment

	for i, p in enumerate(document.paragraphs):
		# Special blocks, alignment
		txt = p.text.strip()
		if txt == '[[SEP-ITEM]]':
			p.text = sep
			p.style = document.styles['SepStyle']
			p.alignment = ALIGMENTS['center']


			p.paragraph_format.space_before = SEP_MARGIN
			p.paragraph_format.space_after = SEP_MARGIN
		elif txt == '[[SKIP-ITEM]]':
			p.text = ''
			p.add_run().add_break(WD_BREAK.PAGE)
		else:
			if p.style.name == heading_sep and i > 0:
				# document.paragraphs[i-1].add_run().add_break(WD_BREAK.PAGE)
				p.paragraph_format.page_break_before = True

			if alignment in ALIGMENTS:
				p.alignment = ALIGMENTS[alignment]

		# Images
		images = []
		def detect_image(match):
			txt, url = match.group(1), match.group(2)
			images.append(url)
			return ''

		if "[[IMAGE-ITEM]]" in p.text:
			p.text = re.sub(REGEX_IMAGES, detect_image, p.text)
			r = p.add_run()
			for url in images:
				url = get_file_local_path(url)
				if url:
					img = r.add_picture(url)
			if image_alignment in ALIGMENTS:
				p.alignment = ALIGMENTS[image_alignment]

	document.save(docx_path)

# -------------------- MAKE MARDOWN-HTML CONVERTABLE -------------------- #

def purify_for_docx(code):
	REGEX_MD_IMAGE = r"!\[(.*?)\]\((.*?)\)"

	def store_html_image(match):
		txt, url = match.group(1), match.group(2)
		template = '[[IMAGE-ITEM]][[{}]][[{}]]'
		return template.format(txt, url)

	code = purify_remove_html(code)
	code = re.sub(REGEX_MD_IMAGE, store_html_image, code)
	return code


from pathlib import Path

from md2book.config import *
from .common import rand_str, is_int
from .exceptions import ConfigError

# -------------------- CSS -------------------- #

def create_css_file(content):
	filepath = str(TMP_DIRS[0] / ("style-" + rand_str(10) + ".css"))
	with open(filepath, 'w') as f:
		f.write(content)
	return filepath

def ensure_with_unit(val, default_unit='px'):
	try:
		return str(int(val)) + default_unit
	except:
		return val

# -------------------- FONTS -------------------- #

class FontFamily:
	HTML_PATH = "{syspath}"
	EPUB_PATH = "../fonts/{file}"

	def __init__(self, name):
		basepath = Path(name)
		if '/' not in name and '\\' not in name:
			basepath = EMBED_FONTS_PATH / name
		name = basepath.name

		self.font_path = basepath.resolve()
		self.name = name

		if not self.font_path.exists() or not self.font_path.is_dir():
			raise ConfigError("Font {} ({}) doesn't exists".format(self.name, str(self.font_path)))

		font_files = [p for p in self.font_path.iterdir() if p.stem.startswith(self.name)]
		self.font_files = []

		for file in font_files:
			try:
				name, style, weight = file.stem.split('-')
			except ValueError as e:
				raise ConfigError('A font file should be name using the format "fontname-style-weight"')
			if style not in ['normal', 'italic']:
				raise ConfigError("{} is not a valid style for a font (first parameter)".format(style))
			if weight not in ['normal', 'bold'] and not is_int(weight):
				raise ConfigError("{} is not a valid weight for a font (second parameter)".format(weight))

			self.font_files.append((file, style, str(weight)))

	def get_css(self, form):
		if form in ['epub']:
			src_template = self.EPUB_PATH
		else:
			src_template = self.HTML_PATH

		css = []
		for file, style, weight in self.font_files:
			src = src_template.format(
				file=file.name,
				syspath=str(file),
			)
			css.append(FONT_FAMILY_CSS.format(
				font=self.name,
				style=style,
				weight=weight,
				src=src
			))
		return css



from md2book.config import *
from .pipelines import *

# -------------------- EXTERNAL FUNCTIONS -------------------- #

FORMATS_PIPELINE = {
	'markdown' : PipeMd,
	'html' : PipeMd2Html,
	'pdf' : PipeMd2Html2Pdf,
	'docx' : PipeMd2DocxPandoc,
	'odt' : PipeMd2OdtPandoc,
	'epub' : PipeMd2EpubPandoc,
	'txt' : PipeMd2txtPandoc,
}

def convertBook(code, target):
	code = MarkdownCode(code)
	pipeline = FORMATS_PIPELINE[target.format]
	code = pipeline(code).execute(target)

	file_name = target['name'] + '.' + code.EXT
	out_file = target.compile_dir / file_name
	code.output(out_file)

	return out_file


import subprocess, os, platform, fnmatch, yaml, time
import random, string
from copy import deepcopy
import ssl
from urllib.request import urlopen
from pathlib import Path

from .exceptions import ConfigError
from md2book.config import *

ctx = ssl.create_default_context()
ctx.check_hostname = False
ctx.verify_mode = ssl.CERT_NONE

# -------------------- UTILITY -------------------- #

def merge_dicts_recur(settings, overrride):
    if overrride is None:
        return settings
    for key, val in overrride.items():
        if isinstance(val, dict) and isinstance(settings.get(key, None), dict):
            merge_dicts_recur(settings[key], val)
        else:
            settings[key] = deepcopy(val)
    return settings

def rand_str(n):
    return "".join(random.choices(string.ascii_lowercase, k=n))

def is_int(n):
    try:
        m = int(n)
        return True
    except:
        return False

# -------------------- SYSYTEM -------------------- #

def sys_open(filepath):
    filepath = str(filepath)
    if platform.system() == 'Darwin':       # macOS
        subprocess.call(('open', filepath))
    elif platform.system() == 'Windows':    # Windows
        os.startfile(filepath)
    else:                                   # linux variants
        subprocess.call(('xdg-open', filepath))

# -------------------- FILES -------------------- #

def escapePath(path):
    path = str(path)
    path = path.replace(" ", "\\ ")
    return path

def find_files_matching(path, pattern):
    if path.is_file():
        if fnmatch.fnmatch(path.name, pattern):
            yield path
    elif path.is_dir():
        for sub_path in path.iterdir():
            yield from find_files_matching(sub_path, pattern)

def load_yaml_file(filepath, default=None):
    filepath = str(filepath)
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            try:
                return yaml.load(f, Loader=yaml.FullLoader)
            except:
                raise ConfigError("Invalid yaml format", filepath)
    except FileNotFoundError:
        if default is None:
            raise ConfigError("This file doesn't exists", filepath)
    return default

def download_url(url, ext='', path=None, repeat_n=2, repeat_delay=0.5):
    if path is None:
        path = str(TMP_DIRS[0] / (rand_str(20) + ext))
    try:
        with urlopen(url) as img:
            with open(path, 'wb') as svg:
                svg.write(img.read())
        return path
    except Exception as e:
        if repeat_n > 0:
            time.sleep(repeat_delay)
            return download_url(url, ext=ext, path=path, repeat_n=repeat_n-1, repeat_delay=repeat_delay*2)
        print('ERROR', e)
        return None

def get_file_in(path, path_list=['/'], dir_ok=False):
    for base in path_list:
        if base:
            complete_path = (Path(base) / path).resolve()
            if complete_path.exists() and (complete_path.is_file() or dir_ok):
                return str(complete_path)
    return download_url(path)
    

def get_file_local_path(path, base_path='', user_path='', dir_ok=False):
    path_list = [base_path, user_path, '/']
    return get_file_in(path, path_list, dir_ok)


import yaml, re
from copy import deepcopy

from md2book.config import *
from .common import merge_dicts_recur, load_yaml_file
from .style import create_css_file
from .exceptions import ConfigError
from md2book.modules import ALL_MODULES

# -------------------- MAIN SETTINGS -------------------- #

def merge_settings(settings, overrride):
	for key, val in overrride.items():
		if isinstance(val, dict) and isinstance(settings.get(key, None), dict):
			merge_settings(settings[key], val)
		else:
			settings[key] = deecopy(val)
	return settings

def load_settings():
	settings = load_yaml_file(DEFAULT_SETTINGS, {})

	# Check the settings
	if not isinstance(settings, dict):
		settings = {}

	def_target = settings.get("default_target", {})

	for key, val in DEFAULT_TARGET.items():
		if isinstance(val, dict) and not isinstance(def_target.get(key, None), dict):
			def_target[key] = {}
		elif isinstance(val, list) and not isinstance(def_target.get(key, None), list):
			def_target[key] = []

	# Merge settings, and save the file
	try:
		with open(GENERATED_SETTINGS_FILE, 'w') as f:
			yaml.dump({'default_target': DEFAULT_TARGET}, f)
	except: # If the user can't write to this location
		pass
	merge_dicts_recur(DEFAULT_TARGET, def_target)

# -------------------- BOOK CONFIG -------------------- #

def create_default_book_config(path, conf_path):
	name = path.with_suffix("").name
	title = re.findall(r"[a-zA-Z0-9']+", name)
	title = " ".join(title).capitalize()

	print("{} doesn't exists and will be created".format(str(conf_path)))
	print("- title =", title)
	print("- file name =", name)

	conf_content = {'targets' : {'main' : {
		**SIMPLE_TARGET, **{
		'chapters' : [str(path.parts[-1])],
		'name' : name,
		'title' : title,
	}}}}
	with open(str(conf_path), 'w') as file:
		yaml.dump(conf_content, file)

# -------------------- BOOK TARGETS -------------------- #

class Target:
	def __init__(self, conf=None, path='.', compile_dir='.'):
		self.conf = conf or deepcopy(DEFAULT_TARGET)
		self.loaded_conf = {}
		self.loaded = set()
		self.path = Path(path)
		self.compile_dir = Path(compile_dir)
		self.is_linked_loaded = False

		self.modules = []
		self.mods = {} # Modules with a name

		self.stylesheets = []
		self.scripts = []

	def __getitem__(self, key):
		return self.conf.get(key)

	def load_from_settings(self, config, form):
		self.conf = {}
		self._load_settings_recur(config, form)
		self.loaded_conf = deepcopy(self.conf)
		self.conf = merge_dicts_recur(deepcopy(DEFAULT_TARGET), self.conf)

	def _load_settings_recur(self, config, form):
		if form in ALLOWED_FORMATS:
			config[form] = config.get(form, {})
			config[form]['format'] = form

		if form in self.loaded:
			return {}
		self.loaded.add(form)

		inherit = ['main'] + config.get('inherit', [])
		for subtarget in inherit:
			self._load_settings_recur(config, subtarget)

		self.merge(config[form])

	def is_user_defined(self, key):
		if isinstance(key, str):
			key = key.split('.')
		key = key[::-1]
		arg = self.loaded_conf
		while key and arg is not None:
			arg = arg.get(key.pop(), None)
		return (arg is not None)

	def merge(self, overrride):
		self.conf = merge_dicts_recur(self.conf, overrride)

	def resolve_format(self):
		self.format = self.conf['format']
		while self.format in FORMAT_ALIASES:
			self.format = FORMAT_ALIASES[self.format]
		if not self.format in BASE_FORMATS:
			raise ConfigError("Format {} doesn't exists".format(self.format))

	def load_modules(self):
		for module_cls in ALL_MODULES:
			if module_cls.NAME is None:
				self.modules.append(module_cls(None, self))
			elif module_cls.NAME in self.conf:
				self.modules.append(module_cls(self.conf.get(module_cls.NAME), self))
				self.mods[module_cls.NAME] = self.modules[-1]
			else:
				self.mods[module_cls.NAME] = None

	def complete(self):
		self.resolve_format()
		self.load_modules()

	def load_linked_files(self):
		custom_css = []
		if self.is_linked_loaded:
			return
		self.is_linked_loaded = True

		for mod in self.modules:
			if mod.get_conf_name():
				self.conf[mod.get_conf_name()] = mod.conf

			self.stylesheets.extend(mod.get_stylesheets())
			self.scripts.extend(mod.get_scripts())
			custom_css.extend(mod.get_custom_css())

		self.stylesheets.append(create_css_file("\n".join(custom_css)))

		for mod in self.modules:
			mod.alter_target()

		self.stylesheets = [str(p) for p in self.stylesheets]



import re

from .mdhtml import purify_remove_html

TXT_LARG = 72
SKIP = '\n' * 25

def center_txt(text):
	if isinstance(text, list):
		return [center_txt(l) for l in text]
	left = max(0, TXT_LARG - len(text)) // 2
	return ' ' * left + text

def purify_for_txt(code):
	REGEX_HTML_LINK = r'<a.*?href=([\'"])(.*?)\1.*?>(.*?)</.*?a.*?>'
	REGEX_HTML_ELEM = r'</?[a-zA-Z].*?>'

	def convert_link(match):
		return '[{text}]({url})'.format(text=match.group(3), url=match.group(2))

	code = purify_remove_html(code)
	code = re.sub(REGEX_HTML_LINK, convert_link, code)
	code = re.sub(REGEX_HTML_ELEM, '', code)

	return code

def post_process_txt(code, target):
	sep = center_txt(target['sep'])

	code = code.replace('\\[', '[')
	code = code.replace('\\]', ']')
	code = code.replace('\\]', ']')

	REGEX_IMAGE = r'!\[(.*?)\]\((.*?)\)'
	REGEX_LINK = r'\[(.*?)\]\((.*?)\)'
	# REGEX_MARKDOWN_EL = r'(\*|\*\*|_|__|`|```|~~|~|\^)(\S.*?)\1'
	REGEX_MARKDOWN_EL_2 = r'(\*\*|__|``|~~)(\S.*?)\1'
	REGEX_MARKDOWN_EL = r'(\*|_|`|~|\^)(\S.*?)\1'

	def convert_image(match):
		text = match.group(1).strip()
		if text:
			text =' (' + text + ')'
		return '[Image{}: {}] '.format(text, match.group(2))
	def convert_link(match):
		return '{} ({})'.format(match.group(1), match.group(2))
	def convert_md_el(match):
		return match.group(2)

	code = re.sub(REGEX_IMAGE, convert_image, code)
	code = re.sub(REGEX_LINK, convert_link, code)
	code = re.sub(REGEX_MARKDOWN_EL_2, convert_md_el, code)
	code = re.sub(REGEX_MARKDOWN_EL, convert_md_el, code)

	code = code.replace('[[SEP-ITEM]]', sep)
	code = code.replace('[[SKIP-ITEM]]', SKIP)

	return code


import datetime, yaml, re
import urllib, hashlib

from .base import BaseModule
from md2book.config import *
from md2book.templates import TemplateFiller
from md2book.formats.mdhtml import extract_toc
from md2book.util.exceptions import SimpleWarning
from md2book.util.common import download_url

class MetadataModule(BaseModule):
	NAME = 'metadata'
	def __init__(self, conf, target):
		super().__init__(conf, target)
		meta = self.conf

		meta['title'] = target.conf['title'].strip()

		if target.conf['subtitle']:
			meta['subtitle'] = target.conf['subtitle'].strip()

		if meta.get('author', None) is None and target.conf['by']:
			meta['author'] = target.conf['by']

		if meta.get('date', None) == 'current':
			meta['date'] = datetime.now().strftime("%d/%m/%Y")

	def get_yaml_intro(self):
		metadata = {key : val for key, val in self.conf.items() if val is not None}
		content = "---\n" + yaml.dump(metadata) + "\n---\n"
		content = content.replace("_", "\\_")
		return content

class ImagesModule(BaseModule):
	NAME = 'images'

	REGEX_RM_IMG = r"<img.*?>|!\[.*?\]\(.*?\)"
	RE_IMG_HTML = r'<img(.*?)/?>'
	RE_EMPTY_ALT = r'alt=([\'"])\1'

	def __init__(self, conf, target):
		super().__init__(conf, target)

		conf['remove'] = bool(conf['remove'])

		if conf['cover']:
			conf['cover'] = str((target.path.parent / conf['cover']).resolve())
			target.conf['metadata']['cover-image'] = conf['cover']

	def replace_html_images(self, match):
		code = match.group(1)
		if not " alt=" in code:
			code = code + ' alt="Image" '
		return '<img{}/>'.format(code)

	def alter_md(self, code):
		if self.conf['remove']:
			code.code = re.sub(self.REGEX_RM_IMG, '', code.code)

	def alter_html(self, code):
		code.code = re.sub(self.RE_IMG_HTML, self.replace_html_images, code.code)
		code.code = re.sub(self.RE_EMPTY_ALT, 'alt="Image"', code.code)

class TitlePageModule(BaseModule):
	NAME = 'titlepage'

	def alter_html(self, code):
		if self.format != 'epub':
			with open(TITLE_PAGE_TEMPLATE, 'r') as f:
				titlepage = f.read()
			titlepage = TemplateFiller(self.target).fill(titlepage)
			code.code = titlepage + code.code

class TocModule(BaseModule):
	NAME = 'toc'

	def pandoc_options(self, dest_format):
		options = super().pandoc_options(dest_format)
		chapter_level = 1
		if self.conf['enable']:
			chapter_level = min(3, max(1, int(self.conf['enable'])))
			options.append("--toc")
			options.append('--toc-depth=' + str(self.conf['level']))

		if dest_format == 'epub':
			options.append("--epub-chapter-level=" + str(chapter_level))
		return options

	def alter_md(self, code):
		if self.format in ['html', 'pdf', 'markdown', 'txt'] and self.conf['enable']:
			code.code = '[TOC]\n\n' + code.code

		if self.format in ['epub'] and '[TOC]' in code.code:
			toc_html = extract_toc(code.code, self.conf['level'], self.format)
			code.code = code.code.replace('[TOC]', toc_html)

		if self.format in ['odt', 'epub', 'docx', 'html_light']:
			if self.conf['enable'] is None:
				self.conf['enable'] = '[TOC]' in code.code
			code.code = code.code.replace('[TOC]', '')

	def get_stylesheets(self):
		styles = []
		style = self.conf['style'].strip()
		if style:
			styles.append(DATA_PATH / 'styles/toc/base.css')
			if style != 'base':
				styles.append(DATA_PATH / 'styles/toc/{}.css'.format(style))
		return styles

class HtmlBlocksModule(BaseModule):
	RE_BR = r'<br(.*?)>'
	REGEX_COMMENTS = r"<!--([\s\S]*?)-->"
	RE_SVG_IMAGE = r'(<img .*?src="(.*?\.svg)".*?>)'

	def __init__(self, conf, target):
		super().__init__(conf, target)
		self.alter_svg = target.format in ['html', 'pdf']

	def alter_html(self, code):
		code.code = re.sub(self.REGEX_COMMENTS, '', code.code)
		code.code = re.sub(self.RE_BR, '<br />', code.code)
		if self.alter_svg:
			code.code = re.sub(self.RE_SVG_IMAGE, self.svg_inserter, code.code)

	def svg_inserter(self, match):
		path = match.group(2)
		if path.startswith('http://') or path.startswith('https://'):
			return math.group(1)
		with open(str(path)) as f:
			xml = f.readlines()[2:]
		xml = ''.join(xml)
		return xml


class LatexModule(BaseModule):
	NAME = 'latex'
	TEX_BLOCKS = r'\$\$([\s\S]*?)\$\$'
	TEX_INLINE = r'\$(.*?)\$'
	SVG_SIZE= r'width="(.*?)" height="(.*?)"'
	# XML_PROP_RE = r'[\s\S]*?<svg.*?width="(.*?)".*?height="(.*?)".*?role="img"[\s\S]*?'
	# IMAGE_TEMPLATE = '![]({url})'
	IMAGE_INLINE = '<img src="{}" style="width:{}; height: {};" />'
	IMAGE_BLOCK = '<p class="centerblock"><img src="{}" style="width:{}; height: {};" /></p>'
	ALIASES = {
		r'\\R' : r'\\mathbb{R}',
		r'\\infin' : r'\\infty',
	}

	def __init__(self, conf, target):
		super().__init__(conf, target)
		self.enabled = bool(self.conf)
		self.download_status = None
		self.equations_names = set()
		self.relative_path = target.format in ['md']

		if self.enabled:
			self.download_dir = target.compile_dir / 'latex'
			self.download_dir.mkdir(parents=True, exist_ok=True)

	def get_latex_image_code(self, path, inline=True):
		with open(str(path)) as f:
			xml = f.readlines()
		xml = ''.join(xml)
		sizes = re.search(self.SVG_SIZE, xml)

		template = self.IMAGE_INLINE if inline else self.IMAGE_BLOCK
		path = ('latex/' + path.name) if self.relative_path else str(path)
		return template.format(path, sizes.group(1), sizes.group(2))

	def clean_cache(self):
		for file in self.download_dir.iterdir():
			if not file.name in self.equations_names:
				file.unlink()

	def preprocess_text(self, texcode):
		a = texcode
		for alias in self.ALIASES:
			texcode = re.sub(alias + r'(?=([^a-zA-Z\d]|$))', self.ALIASES[alias], texcode)
		return texcode

	def get_inserter(self, argname):
		def match_latex(match):
			if self.download_status == 'offline':
				return ''

			texcode = match.group(1).strip()
			texcode = self.preprocess_text(texcode)
			args = urllib.parse.urlencode({'color' : 'black', argname : texcode})
			url = 'https://math.vercel.app?' + args
			filename = hashlib.md5(url.encode('utf-8')).hexdigest()

			path = self.download_dir / (filename + '.svg')
			if not path.resolve().is_file():
				if self.download_status is None:
					print("Download LaTeX images from math.now.sh...")
				result_path = download_url(url, path=str(path))
				if result_path is None:
					self.download_status = 'offline'
					return ''
				else:
					self.download_status = 'ok'

			# path = path.relative_to(self.target.compile_dir)
			path = path.resolve()
			self.equations_names.add(path.name)
			return self.get_latex_image_code(path, inline=(argname == 'inline'))

		return match_latex

	def alter_md(self, code):
		if self.enabled:
			code.code = re.sub(self.TEX_BLOCKS, self.get_inserter('from'), code.code)
			code.code = re.sub(self.TEX_INLINE, self.get_inserter('inline'), code.code)

			if self.download_status == 'offline':
				SimpleWarning('The math.now.sh API is not accessible, LaTeX may not be rendered').show()
			elif self.download_status == 'ok':
				print("LaTeX download complete")
			if self.download_status != 'offline':
				self.clean_cache()


from copy import deepcopy

from md2book.config import *
from md2book.util.common import escapePath, get_file_in
from md2book.util.exceptions import ConfigError

class BaseModule:
	NAME = None # None for a module auto-loaded without configuration
	
	def __init__(self, conf, target):
		self.conf = conf
		self.target = target
		self.format = target.format

	def get_conf_name(self):
		return self.NAME

	def get_custom_css(self):
		return []

	def get_stylesheets(self):
		return []

	def get_scripts(self):
		return []

	def alter_target(self):
		pass

	def alter_html(self, code):
		pass

	def pandoc_options(self, dest_format):
		return []

	def alter_md(self, code):
		pass

# -------------------- MODULE FOR ALL BASE DATAS -------------------- #

class MainTargetDatasModule(BaseModule):
	"""
		This module does not behave like all other modules.
		It manage the BASE configuration
	"""
	def alter_target(self):
		conf = self.target.conf
		if conf.get('title', None) is None:
			conf['title'] = conf['name']

		# Stylesheets
		conf_css = []
		base_paths = ['/', self.target.path.parent, DATA_PATH / 'styles']
		for path in  conf['css']:
			real_path = get_file_in(path, base_paths)
			if real_path:
				conf_css.append(real_path)
			else:
				raise ConfigError("Can't find stylesheet {}".format(path))

		base_styles = deepcopy(BASE_STYLES['default'])
		base_styles.extend(self.target.mods['theme'].styles)
		if self.format in BASE_STYLES:
			base_styles.extend(BASE_STYLES[self.format])

		self.target.stylesheets = base_styles + self.target.stylesheets + conf_css

		# Scripts
		for script in conf['js']:
			if script in AVAILABLE_SCRIPTS:
				script = AVAILABLE_SCRIPTS[script]
			self.target.scripts.append(script)

	def pandoc_options(self, dest_format):
		options = super().pandoc_options(dest_format)
		if dest_format == 'epub':
			for path in self.target.stylesheets:
				options.append("--css=" + escapePath(path))
		return options


import re

def preparse_ebook_markdown(code):
	RREGEX_MARK = r'(==)(.*?)=='

	def replace_mark(match):
		return '<mark>{}</mark>'.format(match.group(1))

	code = re.sub(RREGEX_MARK, replace_mark, code)
	return code


import shutil

from md2book.util.common import rand_str
from md2book.templates import TemplateFiller
from md2book.config import *

# -------------------- CODE CLASSES -------------------- #

class CodeData:
	EXT = "txt"
	REF_DOC = False

	def assertLang(self, *langsCls):
		for lang in langsCls:
			if isinstance(self, lang):
				return True
		raise exception("Code {} is not in {}".format(self.lang, str(langs)))

	def getStorageFile(self):
		filepath = TMP_DIRS[0] / (rand_str(10) + '.' + self.EXT)
		filepath.parent.mkdir(parents=True, exist_ok=True)
		return filepath

	def getStrPath(self):
		return str(self.getStorageFile())

	def getPandocOutputOptions(self):
		return []

	def set_conf(self, target):
		self.target = target
		target.load_linked_files()

	def load_from(self, path):
		raise Exception("Not implemented")

	def output(self, dest=None):
		raise Exception("Not implemented")

	def addStyle(self, path, base_path=""):
		pass

# ----- Pure code

class PureCodeData(CodeData):
	def __init__(self, code=""):
		self.code = code

	def fill_template(self, target):
		self.code = TemplateFiller(target).fill(self.code)

	def get(self):
		return self.code

	def output(self, dest=None):
		if not dest:
			dest = self.getStorageFile()
		with open(str(dest), 'w') as f:
			f.write(self.get())
		return dest

	def load_from(self, path):
		with open(str(path)) as f:
			self.code = f.read()

# ----- File-based

class OutFileCode(CodeData):
	def __init__(self):
		self.filepath = self.getStorageFile()

	def get(self):
		return self.filepath

	def getStrPath(self):
		return str(self.filepath.resolve())

	def output(self, dest=None):
		if dest:
			shutil.copy(str(self.filepath), str(dest))
			return dest
		return self.filepath

	def load_from(self, path):
		self.filepath = path

	def getPandocOutputOptions(self):
		params = super().getPandocOutputOptions()
		if self.REF_DOC:
			params.append('--reference-doc=' + str(DATA_PATH / 'templates' / ('ref.' + self.EXT)))
		for mod in self.target.modules:
			params.extend(mod.pandoc_options(self.EXT))

		return params


from .base import BaseModule
from md2book.config import *
from md2book.util.style import ensure_with_unit, FontFamily
from md2book.util.common import escapePath, load_yaml_file
from md2book.util.exceptions import ConfigError

class StyleModule(BaseModule):
	NAME = 'style'
	def get_stylesheets(self):
		styles = []
		if self.conf.get("center-blocks", None):
			styles.append(DATA_PATH / 'styles' / 'centerblocks.css')

		return styles

	def get_custom_css(self):
		css = []
		chappter_level = self.target.conf['chapter-level']
		if chappter_level:
			css.append("h" + str(chappter_level).strip() + " { page-break-before: always;}")

		if self.conf['indent']:
			indent = ensure_with_unit(self.conf['indent'], 'em')
			css.append('body > p, section > p {text-indent : ' + indent + ';}')

		if self.conf['align']:
			css.append('body {text-align : ' + self.conf['align'] + ';}')

		if self.conf['paragraph-spacing'] is not True:
			spacing = self.conf['paragraph-spacing'] or '0px'
			spacing = ensure_with_unit(spacing, 'px')
			css.append('body > p, section > p {margin : ' + spacing + ' 0px;}')

		return css
	
class SepModule(BaseModule):
	NAME = 'sep'
	def __init__(self, conf, target):
		self.sep = str(conf)
		super().__init__(conf, target)

	def get_custom_css(self):
		sep = self.sep.replace('"', '\\"')
		return [".sepcontent::before{content : \"" + sep + "\"}"]

class FontModule(BaseModule):
	NAME = 'font'
	FONT_CSS = 'body {{ font-family: "{font}","Clear Sans","Helvetica Neue",Helvetica,Arial,sans-serif;}}'

	def __init__(self, conf, target):
		super().__init__(conf, target)
		self.try_add_font(self.conf['default'])

	def get_fonts(self):
		return [FontFamily(font) for font in self.conf['include']]

	def try_add_font(self, name):
		base_font_dir = DATA_PATH / 'fonts' / str(name)
		if name not in self.conf['include'] and base_font_dir.exists():
			self.conf['include'].append(name)

	def get_custom_css(self):
		css = []
		for font in self.get_fonts():
			css.extend(font.get_css(self.format))
		if self.conf['default']:
			css.append(self.FONT_CSS.format(font=self.conf['default']))
		if self.conf['size']:
			size = ensure_with_unit(self.conf['size'], 'px')
			css.append('body { font-size: ' + size + ';}')
		return css

	def pandoc_options(self, dest_format):
		options = super().pandoc_options(dest_format)
		if dest_format == 'epub':
			for font in self.get_fonts():
				path = escapePath(font.font_path / (font.name + "*"))
				options.append("--epub-embed-font=" + path)
		return options

class ThemeModule(BaseModule):
	NAME = 'theme'

	def __init__(self, conf, target):
		super().__init__(conf, target)
		self.theme = conf
		self.styles = []
		self.settings = {}

		if self.theme:
			theme_path = DATA_PATH / 'themes' / (self.theme + '.yml')
			if not (theme_path.exists() and theme_path.is_file()):
				raise ConfigError("The theme {} doesn't exists".format(self.theme))

			self.settings = load_yaml_file(theme_path)
			self.styles = [DATA_PATH / 'styles' / s for s in self.settings.get('css', [])]

			if 'toc' in self.settings and not self.target.is_user_defined('toc.style'):
				self.target.conf['toc']['style'] = self.settings['toc']

			if 'font' in self.settings and not self.target.is_user_defined('font.default'):
				self.target.conf['font']['default'] = self.settings['font']



from .base import *
from .mdfilemods import *
from .stylemods import *

ALL_MODULES = [
	MainTargetDatasModule,
	ThemeModule,
	StyleModule,
	SepModule,
	FontModule,
	MetadataModule,
	LatexModule,
	ImagesModule,
	TitlePageModule,
	TocModule,
	HtmlBlocksModule,
]


from pathlib import Path
import md2book

# -------------------- PATHS -------------------- #

TMP_DIRS = []
SCRIPT_PATH = Path(__file__).parent.resolve()
DATA_PATH = (Path(__file__).parent / 'data').resolve()

GENERATED_SETTINGS_FILE = DATA_PATH / 'generated_default_settings.yml'
EMBED_FONTS_PATH = DATA_PATH / 'fonts'

# In user directory

CONFIG_PATH = (Path.home() / '.md2book').resolve()
DEFAULT_SETTINGS = CONFIG_PATH / 'settings.yml'

try:
	CONFIG_PATH.mkdir(parents=True, exist_ok=True)
	if not DEFAULT_SETTINGS.exists():
		DEFAULT_SETTINGS.touch()
except:
	pass

# -------------------- MAIN -------------------- #

M2B_NAME = 'md2book'
M2B_EMAIL = 'webalorn@gmail.com'
M2B_SHORT_DESCRIPTION = 'md2book (version {}) : Compile markdown book with a simple \
command and a configuration file. - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  \
Documentation available at https://github.com/webalorn/md2book/wiki - - - - -  \
Repport bugs on https://github.com/webalorn/md2book/issues -- - - - - - - - -  \
'.format(md2book.__version__)
M2B_EPILOG = 'Settings : {}\
'.format(str(DEFAULT_SETTINGS))


# -------------------- TARGETS -------------------- #

BOOK_FILE_NAMES = "*book.yml"
DEFAULT_GEN_DIR = "generated"

BASE_STYLES = {
	"default" : ["default.css"],
	"html" : ["html.css"],
	"pdf" : ["pdf.css"],
	"epub" : ["epub.css"],
}
BASE_STYLES = {key : [str(DATA_PATH / "styles" / p) for p in val] for key, val in BASE_STYLES.items()}

DEFAULT_TARGET = {
	'inherit' : [],
	'chapters' : [],
	'variables' : {},

	'name' : 'book', # Name of the file created
	'title' : None, # Displayed name or title. Default is the name
	'subtitle' : None, # Displayed name or title. Default is the name
	'by' : None,

	'format' : 'pdf',

	'theme' : 'github', # Set 'no' for no theme.
	'css' : [],
	'js' : [],
	'between-chapters' : '\n\n',
	'chapter-level' : 1, # or 2, 3 : level where to split into chapters

	# Modules : 
	'sep' : '✶   ✶   ✶',
	'latex' : True,
	'titlepage' : {
		'enable' : True,
		'image' : None,
	},
	'images' : {
		'remove' : False,
		'cover' : None,
	},
	'font': {
		'default' : 'opensans',
		'include' : [], # Additional fonts
		'size' : None,
	},
	'toc' : {
		'enable' : True, # Generate a TOC at the begening of the file
		'level' : 3,
		'style' : 'base',
	},
	'style' : {
		'align' : 'justify',
		'center-blocks' : True,
		'indent' : False,
		'paragraph-spacing' : True,
	},
	'metadata' : {
		# 'title' : [], # Title and subtitle from the config. You can add short, collection, edition, extended
		# 'subtitle' : "",
		'author' : None, # [ author1, author2, ...]
		'keywords' : None, # [...]
		'abstract' : None, # Text
		'date' : None, # current will be replaced by the current date. Otherwise, None, or YYYY-MM-DD.
		'lang' : None,
		'subject' : None, # [...]
		'description' : None, # Or text
		'rights' : None,
		'thanks' : None,
	}
}

SIMPLE_TARGET = { # Field in a generated configuration file
	'name' : 'book',
	'title' : None,
	'subtitle' : None,
	'by' : None,
	'theme' : 'github',

	'font' : {
		'default' : 'opensans',
		'size' : None,

	},
}

# -------------------- TEMPLATES -------------------- #

HTML_TEMPLATE = DATA_PATH / 'templates' / 'structure.html'
TITLE_PAGE_TEMPLATE = DATA_PATH / 'templates' / 'titlepage.html'
FONT_FAMILY_CSS = """
@font-face {{
  font-family: "{font}";
  font-style: {style};
  font-weight: {weight};
  font-display: swap;
  src: url("{src}");
}}"""

AVAILABLE_SCRIPTS = {
	'jquery' : "https://code.jquery.com/jquery-3.5.1.min.js",
}

# -------------------- CONVERTING -------------------- #

BASE_FORMATS = ['markdown', 'html', 'pdf', 'docx', 'odt', 'epub', 'txt']
FORMAT_ALIASES = {
	'word' : 'docx',
	'libreoffice' : 'odt',
	'ebook' : 'epub',
	'md' : 'markdown',
	'web' : 'html',
	'text' : 'txt',
}
ALLOWED_FORMATS = list(BASE_FORMATS) + list(FORMAT_ALIASES)

def get_markdown_default_extensions():
	# Import is done here to keep minimal depencies for this file
	from md2book.formats.mdhtml import MarkdownExtended, TasklistExtension
	return [
		'extra', 'nl2br', 'sane_lists', 'smarty', 'toc',
		MarkdownExtended(), TasklistExtension()
	]


MD_CONFIG = {
	'toc' : {
		'toc_depth' : 6,
	}
}

PDF_OPTIONS = { # https://wkhtmltopdf.org/usage/wkhtmltopdf.txt
	'margin-top': '0.75in',
	'margin-right': '0.7in',
	'margin-bottom': '0.7in',
	'margin-left': '0.75in',
	'encoding': "UTF-8",
	'quiet': '',
	'enable-local-file-access': '',
  'no-outline': None,
}


import re
from pathlib import Path
from urllib.request import urlopen

from .forms import PureCodeData
from md2book.util.exceptions import ParsingError
from md2book.util.common import get_file_local_path
from md2book.config import *
from md2book.formats.mddocx import purify_for_docx
from md2book.formats.mdtxt import purify_for_txt

class MarkdownCode(PureCodeData):
	EXT = "md"
	
	def __init__(self, code=""):
		self.code = code
		self.meta_code = ""

	def set_conf(self, target):
		self.fill_template(target)
		if target.mods['metadata']:
			self.meta_code = target.mods['metadata'].get_yaml_intro()
		super().set_conf(target)

	def clear_for(self, target_format):
		""" Remove parts that are not needed in some documents """
		if target_format in ['docx']:
			self.code = purify_for_docx(self.code)
		elif target_format in ['txt']:
			self.code = purify_for_txt(self.code)

	def get(self):
		return self.meta_code + self.code

	def get_base_code_only(self):
		return self.code

class TxtCode(PureCodeData):
	EXT = "txt"

class HtmlCode(PureCodeData):
	EXT = "html"
	def __init__(self, code, title, headers=None):
		self.code = code
		self.title = title
		self.headers = headers or []

	def set_conf(self, target):
		super().set_conf(target)
		
		for style in target.stylesheets:
			self.addStyle(style)
		for script in target.scripts:
			self.addScript(script)

	def getHtmlTag(self, tag, content, **keys_vals):
		keys = ''.join([' {}="{}"'.format(key, val) for key, val in keys_vals.items()])
		if content == None:
			htmltag = '<{tag}{keys} />'
		else:
			htmltag = '<{tag}{keys}>{content}</{tag}>'
		return htmltag.format(tag=tag, content=str(content), keys=keys)

	def addHeader(self, tag, content, **keys_vals):
		self.headers.append(self.getHtmlTag(tag, content, **keys_vals))

	def addStyle(self, path, base_path=""):
		base_path = Path(base_path or self.target.path.parent)
		real_path = get_file_local_path(path, base_path=base_path)
		
		if real_path is None:
			raise ParsingError('Can\'t find the stylesheet {}'.format(str(path)))
		with open(real_path) as f:
			self.addHeader('style', f.read())

	def addScript(self, path, base_path=""):
		base_path = Path(base_path or self.target.path.parent)
		real_path = get_file_local_path(path, base_path=base_path)

		if real_path is None:
			raise ParsingError('Can\'t find the script {}'.format(str(path)))
		with open(real_path) as f:
			self.addHeader('script', f.read())

	def get(self):
		with open(HTML_TEMPLATE) as f:
			html_struct = f.read()
		code = html_struct.format(
			headers="\n".join(self.headers),
			title=self.title,
			body=self.code
		)
		return code


import string
from copy import deepcopy as copy
from datetime import datetime

from md2book.util.exceptions import TemplateError
from md2book.util.common import is_int, rand_str

FONT_CSS_TEMPLATE = "\
<style>\
#{id} {{ display : none; }}\
#{id} ~ * {{ {prop} : {val}; }}\
</style>\
<div id='{id}'></div>"
CSS_FONT_PROPERTIES = {
	'font' : 'font-family',
	'color' : 'color',
	'size' : 'font-size',
}

class TemplateEngine(string.Formatter):
	"""
		Template engine : replace things like {% ... %}
			- "?" indicates an optional parameter
			- <...> indicates an arbitrary value <=int>, <=str>, <=function> if there is a type

		Syntax :
			{%<var_name>%} -> python default behavior, prints the variable
			{%<fct=function>:call:?param1:?...%} -> call the function 'fct'
			{%<var_name>:set:<value>%} -> set the variable "var_name" with "value"
			{%<var_name>:counter:?start-value%} -> Print the counter value and add 1.
					If is doesnt exists, create a variable containing 1, or "start-value"
			{%<file_path>:include%} -> Include the file located at file_path

			{%<var_name>:get%} Try to guess the type and return the value with a good type
			{%?<var_name>:eval:<expr>%} -> eval the expression, store it if <var_name> is non-empty. Otherwise, return the value
			{%?<var_name>:if:<cond>:<expr>%}, {{<var_name>:if:<cond>:<expr>:else:<expr>%}
			{%<rel_path>:relpath%} -> Path relative to the target directory
			{%<font-name>:font%} -> Change the font used
				{%<font-name>:font%}, {%<font-size>:size%}, {%<color>:color%} [SIZE MUST HAVE A UNIT]
			{%<var-name>?:format:<code>%} -> In <code> : replace newlines by <br/>
		
	"""
	def __init__(self, values, path):
		super().__init__()
		self.values = copy(values)
		self.path = path.resolve()

	def vformat(self, format_string, args, kwargs):
		used_args = set()
		result, _ = self._vformat(format_string, args, kwargs, used_args, 200)
		self.check_unused_args(used_args, args, kwargs)
		return result

	def get_field(self, field_name, args, kwargs):
		if is_int(field_name):
			return '', ''
		field_name = field_name.strip()
		return field_name, field_name

	def get_value_of(self, val):
		d = self.values
		val_tab = val.lower().split('.')[::-1]
		while val_tab:
			if 'variables' in d and val_tab[-1] in d['variables']:
				d = d['variables'][val_tab.pop()]
			elif val_tab[-1] in d:
				d = d[val_tab.pop()]
			else:
				raise TemplateError("Variable does not exists: " + repr(val))
		return d

	def get_path_file_content(self, path):
		sys_path = self.path / path
		try:
			content = open(str(sys_path), "r").read()
		except FileNotFoundError:
			raise TemplateError("The path {} doesn't exists".format(str(path)))
		content = self.format(content)
		return content

	def end_expr(self, val, result):
		result = str(result)
		if val:
			self.values[val] = result
			return ''
		return result

	def guess_type_convert(self, val):
		if not isinstance(val, str):
			return str(val)
		v = val.strip()
		if v in ['True', 'true', 'yes']:
			return 'True'
		if v in ['False', 'false', 'no', 'None', 'null']:
			return 'False'
		if v in ['None', 'null', 'none']:
			return 'None'
		if v.startswith("[") and v.endswith("]"):
			return v
		try:
			return str(int(v))
		except:
			pass
		try:
			return str(float(v))
		except:
			pass
		return '"{}"'.format(val)

	def add_font(self, font):
		pass

	def set_font(self, val, prop):
		if not prop in CSS_FONT_PROPERTIES:
			raise TemplateError("The font property '{}' doesn't exists".format(prop))
		if prop == 'font' and val:
			self.add_font(val)
		prop = CSS_FONT_PROPERTIES[prop]

		val = val or 'inherit'
		block_id = rand_str(15)
		return FONT_CSS_TEMPLATE.format(val=val, id=block_id, prop=prop)

	def split_delimiters(self, params, delims=None, ensure_size=None):
		if not isinstance(delims, list):
			delims = [delims]
		parts = [[]]
		for p in params[::-1]:
			if p in delims:
				if ensure_size is None or len(parts) < ensure_size:
					parts.append([])
			else:
				parts[-1].append(p)
		while len(parts) < ensure_size:
			parts.append([])
		return [':'.join(l) for l in parts]

	def format_html(self, code):
		code = code.replace("\n", "\n<br/>")
		return code

	def format_field(self, val, spec):
		params = [s for s in spec.split(':')][::-1]
		action = params.pop().lower().strip() if params else ''
		
		if action == 'call':
			return str(self.get_value_of(val)(*params))

		elif action == 'set':
			self.values[val] = params.pop().strip()

		elif action == 'get':
			return str(self.guess_type_convert(self.get_value_of(val)))

		elif action == 'counter':
			if val not in self.values:
				self.values[val] = int(params.pop())-1 if params else 0
			self.values[val] = int(self.values[val]) + 1
			return str(self.values[val])

		elif action == 'include':
			return self.get_path_file_content(val)

		elif action == 'format':
			txt = ':'.join(params)
			return self.end_expr(val, self.format_html(txt))

		elif action in ['font', 'color', 'size']:
			return self.set_font(val, action)

		elif action == 'relpath':
			path = self.get_value_of(val)
			if path is None:
				return ''
			return str(self.path / path)

		elif action == 'eval':
			result = eval(':'.join(params))
			return self.end_expr(val, result)

		elif action == 'if':
			condition = params.pop().strip()
			if_expr, else_expr = self.split_delimiters(params, 'else', 2)

			cond = eval(condition, self.values)
			if isinstance(cond, str):
				cond = cond.strip()
			return self.end_expr(val, if_expr if cond else else_expr)

		elif not spec: # Simple variable
			v = self.get_value_of(val)
			if isinstance(v, list):
				v = ", ".join([str(e) for e in v])
			return str(v)
		
		else:
			raise TemplateError("Invalid template : {{{}}}".format(val+":"+spec))
		return ''

	def format(self, txt):
		txt = txt.replace("{%", "𓀀").replace("%}", "𓀁")
		txt = txt.replace("{", "𓂴").replace("}", "𓂶")
		txt = txt.replace("𓀀", "{").replace("𓀁", "}")
		self.openedFont = None

		txt = super().format(txt)

		# if self.openedFont:
		# 	txt += '\n\n</div>'

		txt = txt.replace("𓂴", "{").replace("𓂶", "}")
		return txt

class TemplateFiller(TemplateEngine):
	def add_font(self, font):
		self.target.mods['font'].try_add_font(font)

	def __init__(self, target):
		values = copy(target.conf)
		self.target = target

		now = datetime.now()
		values['date'] = now.strftime("%d/%m/%Y")
		values['datetime'] = now.strftime("%d/%m/%Y %H:%M:%S")
		values['time'] = now.strftime("%H:%M:%S")
		values['hour'] = now.strftime("%H:%M")

		sep_string = values['sep']

		values['skip'] = '<div class="pageBreak"></div>'
		values['sep'] = '<div class="sep"><span>' + sep_string + '</span></div>'

		if target.format in ['epub']:
			if '*' in sep_string or '_' in sep_string:
				values['sep'] = '<div class="sep"><span class="sepcontent"></span></div>'

		if target.format in ['docx', 'txt']:
			values['skip'] = '[[SKIP-ITEM]]'
			values['sep'] = '[[SEP-ITEM]]'

		super().__init__(values, target.path.parent)

	def fill(self, code):
		return self.format(code)


from pathlib import Path
import re

import markdown
from markdown.treeprocessors import Treeprocessor
from markdown.inlinepatterns import SimpleTagPattern
from markdown.extensions import Extension

from md2book.imports.tasklist import TasklistExtension

# -------------------- EXTEND MARKDOWN SYNTAX -------------------- #

class MarkdownExtended(Extension):
	REGEX_DEL = r'(~~)(\S.*?)~~'
	REGEX_MARK = r'(==)(\S.*?)=='
	REGEX_SUP = r'(\^)(\S.*?)\^'
	REGEX_SUB = r'(~)(\S.*?)~'
	# REGEX_SKIP = r'\[SKIP\]'

	def extendMarkdown(self, md):
		del_tag = SimpleTagPattern(self.REGEX_DEL, 'del')
		md.inlinePatterns.register(del_tag, 'del', 15)

		mark_tag = SimpleTagPattern(self.REGEX_MARK, 'mark')
		md.inlinePatterns.register(mark_tag, 'mark', 15)

		sup_tag = SimpleTagPattern(self.REGEX_SUP, 'sup')
		md.inlinePatterns.register(sup_tag, 'sup', 10)

		sub_tag = SimpleTagPattern(self.REGEX_SUB, 'sub')
		md.inlinePatterns.register(sub_tag, 'sub', 10)

# -------------------- EXTRACT TOC FROM MARKDOWN CODE -------------------- #

def extract_toc(code, depth, dest):
	lines = code.split('\n')
	lines = [l for l in lines if l.startswith('#')] # Keep headers
	code = '\n\n'.join(['[TOC]'] + lines)

	html = markdown.markdown(code,
		extensions=['toc'],
		extension_configs={ 'toc' : { 'toc_depth' : depth } }
	)
	lines = [l for l in html.split('\n') if not l.startswith('<h')]
	lines[0] = '<div class="toc toc-in-{dest}">'.format(dest=dest)
	html = '\n'.join(lines)
	return html

# -------------------- PRE-PROCESS MARKDOWN -------------------- #

REGEX_IMG_MD = r'!\[(.*?)\]\((.*?)\)'
REGEX_IMG_HTML1 = r'<img(.*?)src="(.*?)"'
REGEX_IMG_HTML2 = r"<img(.*?)src='(.*?)'"

def md_make_paths_absolute(content, dir_path):
	def complete_path(path):
		if not Path(path).is_absolute() and (dir_path / path).exists():
			return str(dir_path / path)
		return path
		
	def replace_md_images(match):
		return "![{}]({})".format(match.group(1), complete_path(match.group(2)))
	def replace_html_images1(match):
		return '<img{}src="{}"'.format(match.group(1), complete_path(match.group(2)))
	def replace_html_images2(match):
		return '<img{}src="{}"'.format(match.group(1), complete_path(match.group(2)))

	content = re.sub(REGEX_IMG_MD, replace_md_images, content)
	content = re.sub(REGEX_IMG_HTML1, replace_html_images1, content)
	content = re.sub(REGEX_IMG_HTML2, replace_html_images2, content)
	return content

# -------------------- MAKE MARDOWN-HTML CONVERTABLE -------------------- #

def purify_remove_html(code):
	REGEX_IMG_HTML = r"<img(.*?)src=['\"](.*?)['\"](.*?)>"
	REGEX_U = r"==(.*?)=="
	REGEX_COMMENTS = r"<!--([\s\S]*?)-->"
	REGEX_STYLE = r"<style.*?>([\s\S]*?)</style.*?>"

	def replace_html_image(match):
		attributes = match.group(1) + " " + match.group(3)
		return '![]({})'.format(match.group(2))

	def replace_u(match):
		return match.group(1)

	code = re.sub(REGEX_IMG_HTML, replace_html_image, code)
	code = re.sub(REGEX_U, replace_u, code)
	code = re.sub(REGEX_COMMENTS, '', code)
	code = re.sub(REGEX_STYLE, '', code)
	return code


class BaseError(Exception):
	COLORS = {
		'red' : '\u001b[31m',
		'magenta' : '\u001b[35m',
		'default' : '\u001b[0m',
		'white' : '\u001b[97m',
		'green' : '\u001b[92m',
		'yellow' : '\u001b[93m',
	}
	TEMPLATE = "{red}[ERROR]"

	def __init__(self, **variables):
		self.variables = {key : str(v) for key, v in variables.items()}
		super().__init__(self.TEMPLATE)

	def format(self, **variables):
		variables = {**variables, **self.COLORS}
		template = self.TEMPLATE + "{default}"
		return template.format(**variables)

	def __str__(self):
		return self.format(**self.variables)


# -------------------- CUSTOM EXCEPTIONS -------------------- #

class ModuleImportError(BaseError):
	TEMPLATE = "{red}[ERROR] The following python module is not installed : {name}"

class MissingDependencyError(BaseError):
	TEMPLATE = "{red}[ERROR] The following command is required but has not been found : {name}"

class TemplateError(BaseError):
	TEMPLATE = "{red}[ERROR IN TEMPLATE] {white}{error}"
	def __init__(self, error):
		super().__init__(error=error)

# Located errors

class LocatedError(BaseError):
	def set_location(self, path):
		if not self.variables.get('where', ''):
			self.variables['where'] = str(path)

class ConfigError(LocatedError):
	TEMPLATE = "{red}[ERROR] {magenta}{where} {white}: {error}"

	def __init__(self, error, where=''):
		super().__init__(where=where, error=error)

class ParsingError(LocatedError):
	TEMPLATE =  "{red}[ERROR] {magenta}{where} {white} {error}"
	def __init__(self, error):
		super().__init__(error=error, where='')
	
# -------------------- WARNING -------------------- #

class BaseWarning(BaseError):
	def show(self):
		print(str(self))

class SimpleWarning(BaseWarning):
	TEMPLATE = "{yellow}[WARNING] {path} : {message}"

	def __init__(self, message, path=''):
		super().__init__(message=str(message), path=str(path))

class WarningNoBookFound(BaseWarning):
	TEMPLATE = "{yellow}[WARNING] No book found under: {white}{path}"

	def __init__(self, path):
		super().__init__(path=str(path))


import importlib, traceback
# from shutil import which
import distutils.spawn

def which(name):
	return distutils.spawn.find_executable(name)

from md2book.util.exceptions import BaseError, ModuleImportError, MissingDependencyError

dependencies = ['pandoc', 'wkhtmltopdf']
python_modules = {
	'pyyaml' : 'yaml',
	'markdown' : 'markdown',
	'pdfkit' : 'pdfkit',
	'python-docx' : 'docx',
}

def check_dependency(name):
	if which(name) is None:
		raise MissingDependencyError(name=name)

def check_module(name, module):
	if not importlib.util.find_spec(module):
		raise ModuleImportError(name=name)

def main():
	try:
		for name in dependencies:
			check_dependency(name)
		for name, module in python_modules.items():
			check_module(name, module)

		# Import here to load modules AFTER checkin dependencies
		from md2book.main import main
		main()
	except BaseError as e:
		print(e)
	except Exception as e:
		traceback.print_exc()
		print("\u001b[31m[ERROR] An unexpected exception occured in md2book. If you can't solve the problem, please open an issue on https://github.com/webalorn/md2book/issues\nIf possible, copy the error message, the configuration file and the markdown code that created the problem.\u001b[0m")

if __name__ == '__main__':
	main()


import setuptools, os
from pathlib import Path
import importlib.util

cur_dir = Path(__file__).parent
FILES_BLACK_LIST = ['data/default_settings.yml']

try:
	with open("readme.md", "r") as readme_file:
		readme_content = readme_file.read()
except:
	readme_content = '[error] readme.md unavailable'

def find_data_files(root_path, parent_dir):
	parent_dir = (cur_dir / parent_dir).resolve()
	files = []
	def find_sub_data(path):
		if path.is_dir():
			for p in path.iterdir():
				find_sub_data(p)
		elif not path.name.startswith('.'):
			files.append(path.resolve().relative_to(parent_dir))
	find_sub_data(cur_dir / root_path)

	files = [str(f) for f in files]
	files = [f for f in files if f not in FILES_BLACK_LIST]
	return files

def import_module_file(name, path):
	path = str(cur_dir / path)
	spec = importlib.util.spec_from_file_location(name, path)
	mod = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(mod)
	return mod

md2book_module = import_module_file("m2b", "md2book/__init__.py")
md2book_conf = import_module_file("conf", "md2book/config.py")

setuptools.setup(
	name=md2book_conf.M2B_NAME,
	version=md2book_module.__version__,
	author=md2book_module.__author__,
	author_email=md2book_conf.M2B_EMAIL,
	description="Convert markdown files to beautiful books using a simple configuration file",
	long_description=readme_content,
	long_description_content_type="text/markdown",
	url="https://github.com/webalorn/md2book",
	packages=setuptools.find_packages(),
	entry_points={
		'console_scripts': [
			'md2book=md2book.start:main',
		],
	},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	project_urls={
		"Documentation": "https://github.com/webalorn/md2book/wiki",
		"Source": "https://github.com/webalorn/md2book",
	},
	include_package_data=True,
 	package_data={'md2book': find_data_files('md2book/data', 'md2book')},
	install_requires=['pyyaml>=5.2', 'markdown>=3.2', 'pdfkit>=0.6.1', 'python-docx>-0.8.10'],
	python_requires='>=3',
)


import pdfkit, markdown, subprocess
from copy import deepcopy

from md2book.config import *
from md2book.util.exceptions import ParsingError
from md2book.formats.mddocx import post_process_docx
from md2book.formats.mdtxt import post_process_txt
from md2book.formats.mdebook import preparse_ebook_markdown
from .forms_text import *
from .forms_stored import *

# -------------------- META-PIPELINES -------------------- #

class ConvertPipeline:
	BASE_LANG = MarkdownCode
	def __init__(self, code):
		self.code = code

	def alter_code(self, target):
		self.code.clear_for(target.format)
		for mod in target.modules:
			mod.alter_md(self.code)

	def execute(self, target):
		if self.BASE_LANG:
			self.code.assertLang(self.BASE_LANG)
		self.code.set_conf(target)
		self.alter_code(target)

		self.prepare_config(target)
		self.convert_steps(target)

		return self.code

	# Functions to overrride

	def prepare_config(self, target):
		pass

	def convert_steps(self, target):
		pass

# -------------------- PYTHON PIPELINES -------------------- #

class PipeMd(ConvertPipeline):
	pass

class PipeMd2Html(ConvertPipeline):
	def prepare_config(self, target):
		self.md_extensions_conf = deepcopy(MD_CONFIG)
		self.md_extensions = get_markdown_default_extensions()
		self.md_extensions_conf['toc']['toc_depth'] = target['toc']['level']

	def md2html(self, target):
		# First we transform the markdown into html
		html = markdown.markdown(self.code.get_base_code_only(),
			extensions=self.md_extensions,
			extension_configs=self.md_extensions_conf)

		self.code = HtmlCode(html, target['title'])

		for mod in target.modules:
			mod.alter_html(self.code)
		self.code.set_conf(target)

	def convert_steps(self, target):
		self.md2html(target)

class PipeMd2Html2Pdf(PipeMd2Html):
	def prepare_config(self, target):
		super().prepare_config(target)

		self.pdfkit_options = deepcopy(PDF_OPTIONS)
		self.pdfkit_options['title'] = target['title']

	def html2pdf(self, target):
		self.code.assertLang(HtmlCode)
		pdf_code = PdfFileCode()

		pdfkit.from_string(
			self.code.get(),
			pdf_code.getStrPath(),
			options=self.pdfkit_options
		)
		self.code = pdf_code

	def convert_steps(self, target):
		super().convert_steps(target)
		self.html2pdf(target)

# -------------------- PANDOC PIPELINES -------------------- #

class PandocPipeline(ConvertPipeline):
	DEST_FORMAT = MarkdownCode

	def pandoc(self, dest, *add_params):
		origin_out = self.code.output()
		dest_out = dest.output()
		add_params = dest.getPandocOutputOptions() + list(add_params)

		r = subprocess.run(['pandoc', origin_out, '-o', dest_out] + add_params)
		if r.returncode:
			raise ParsingError("Error while converting with pandoc : {}".format(r.stderr))

		dest.load_from(dest_out)
		self.code = dest

	def convert_steps(self, target):
		dest = self.DEST_FORMAT()
		dest.set_conf(target)
		self.pandoc(dest)

class PipeAny2txtPandoc(PandocPipeline):
	BASE_LANG = None
	DEST_FORMAT = TxtCode

class PipeMd2DocxPandoc(PandocPipeline):
	DEST_FORMAT = DocxFileCode

	def convert_steps(self, target):
		super().convert_steps(target)
		post_process_docx(self.code, target)

class PipeMd2OdtPandoc(PandocPipeline):
	DEST_FORMAT = OdtFileCode

class PipeMd2EpubPandoc(PandocPipeline):
	DEST_FORMAT = EpubFileCode

	def alter_code(self, target):
		super().alter_code(target)
		# For markdown not supported by pandoc
		self.code.code = preparse_ebook_markdown(self.code.code)

		# The markdown can contains some HTML that needs to be altered
		for mod in target.modules:
			mod.alter_html(self.code)

class PipeMd2txtPandoc(PandocPipeline):
	DEST_FORMAT = TxtCode

	def convert_steps(self, target):
		super().convert_steps(target)
		self.code.code = post_process_txt(self.code.code, target)

class PipeMd2htmlPandoc(PandocPipeline):
	DEST_FORMAT = HtmlCode

	def md2html(self, target):
		# First we transform the markdown into html
		html = markdown.markdown(self.code.get_base_code_only(),
			extensions=self.md_extensions,
			extension_configs=self.md_extensions_conf)

		self.code = HtmlCode(html, target['title'])

		for mod in target.modules:
			mod.alter_html(self.code)
		self.code.set_conf(target)

	def convert_steps(self, target):
		dest = self.DEST_FORMAT('', target['title'])
		dest.set_conf(target)
		self.pandoc(dest)

		for mod in target.modules:
			mod.alter_html(self.code)
		self.code.set_conf(target)


import argparse, tempfile, glob
from pathlib import Path

from md2book.config import *
from md2book.formats.mdhtml import md_make_paths_absolute
from md2book.util.common import sys_open, load_yaml_file, find_files_matching
from md2book.util.settings import Target, create_default_book_config, load_settings
from md2book.util.exceptions import LocatedError, ConfigError, WarningNoBookFound, SimpleWarning
from md2book.convert.convert import convertBook

def get_cmd_args():
	parser = argparse.ArgumentParser(
		prog=M2B_NAME,
		description=M2B_SHORT_DESCRIPTION,
		epilog=M2B_EPILOG,
	)

	targets = " | ".join(list(ALLOWED_FORMATS))
	parser.add_argument('-t', '--target', action='append', type=str, help='Target to be compiled [main, defined in book.yml, or {}]'.format(targets), default=[])
	parser.add_argument('-o', '--output', type=str, help='Output directory', default=None)
	parser.add_argument('--open', help='Open newly created files', action='store_true', default=False)
	parser.add_argument('--remove-images', help='Remove all images in the document', action='store_true', default=False)
	parser.add_argument('-a', '--all', help='Build all targets defined in the "all" field in the configuration file', action='store_true', default=False)

	parser.add_argument('path', nargs='?', type=str, help='Path from where to search for the books (optional, the default path is the current directory). Can also be a book.yml file or a markdown file.', default='.')

	return parser.parse_args()

# -------------------- BOOKS -------------------- #

def get_real_book_path(path):
	path = Path(path)
	if path.is_file() and path.name.endswith(".md") and path.exists():
		conf = path.with_suffix(".book.yml")
		print("Markdown file detected. The configuration file {} will be used instead".format(str(conf)))
		if not conf.exists():
			create_default_book_config(path, conf)
		return conf
	return path

def find_all_books(path):
	path = Path(path)
	return list(find_files_matching(path, BOOK_FILE_NAMES))

def get_target_md_code(target):
	path = target.path.resolve().parent
	chapters = []
	for chap_name in target['chapters']:
		chap_path = str(path / chap_name)
		files = [chap_path]
		if not Path(chap_path).exists():
			files = glob.glob(chap_path)
			if not files:
				raise ConfigError("Chapter(s) {} not found(s)".format(chap_name))
			# TODO : sort

		for file_path in files:
			with open(file_path, "r", encoding="utf-8") as f:
				content = f.read().strip()

			content = md_make_paths_absolute(content, (path / chap_name).parent.resolve())
			chapters.append(content)
	md_code = target['between-chapters'].join(chapters)
	return md_code

def get_all_targets(book_path):
	config = load_yaml_file(book_path)
	all_targets = config.get('all', None)

	if all_targets is None or not isinstance(all_targets, list):
		raise ConfigError(
			'The field "all" must be defined in the configuration file and be a list to use -a or --all',
			book_path
		)
	if all_targets == []:
		raise SimpleWarning('The "all" field is empty', book_path)

	return [str(t) for t in all_targets]

# -------------------- COMPILE -------------------- #

def check_targets_validity(targets_dict, target_name):
	for t in targets_dict:
		for key in targets_dict[t]: # This can help if a key is mispelled
			if key not in DEFAULT_TARGET:
				raise ConfigError('The configuration field "{}" doesn\'t exists (target "{}")'.format(key, t))
	if 'main' not in targets_dict:
		raise ConfigError('The target \'main\' must be defined')
	if target_name not in list(targets_dict) + ALLOWED_FORMATS:
		raise ConfigError('The target {} is to be built but is not defined'.format(target_name))

def compile_book(book_path, target_name='main', output_dir=None, overwrite_target={}):
	config = load_yaml_file(book_path)
	
	# Extract informations from the config

	if output_dir:
		compile_dir = Path(output_dir)
	else:
		compile_dir = book_path.parent / config.get('compile_in', DEFAULT_GEN_DIR)
	compile_dir.mkdir(parents=True, exist_ok=True)

	targets_dict = config.get('targets', {})
	check_targets_validity(targets_dict, target_name)

	# Load the target
	target = Target(path=book_path, compile_dir=compile_dir, conf={})
	target.load_from_settings(targets_dict, target_name)
	target.merge(overwrite_target)
	target.complete()

	print('========== Compile {} to {}'.format(str(book_path), target.format))
	
	code = get_target_md_code(target)
	out_file = convertBook(code, target)

	print('Success : ', out_file)
	return out_file

# -------------------- MAIN -------------------- #

def main():
	args = get_cmd_args()
	with tempfile.TemporaryDirectory() as tmpdirname:
		TMP_DIRS.append(Path(tmpdirname).resolve())

		load_settings()
		path = get_real_book_path(args.path)
		books = find_all_books(path)
		target_list = args.target or ['main']

		if not books:
			raise WarningNoBookFound(path)

		overwrite_target = {}
		if args.remove_images:
			overwrite_target['remove-images'] = True

		for book in books:
			if args.all:
				target_list = get_all_targets(book)
				print('=> Compile all targets for {} : {}'.format(str(book), ', '.join(target_list)))

			for target in target_list:
				try:
					filepath = compile_book(book, target, args.output, overwrite_target)
					if args.open:
						sys_open(filepath)
				except LocatedError as e:
					e.set_location(book)
					raise e


from .forms import OutFileCode
from md2book.util.common import escapePath

class PdfFileCode(OutFileCode):
	EXT = "pdf"

class DocxFileCode(OutFileCode):
	EXT = "docx"
	REF_DOC = True

class OdtFileCode(OutFileCode):
	EXT = "odt"
	REF_DOC = True

class EpubFileCode(OutFileCode):
	EXT = "epub"